
from __future__ import division
import sys,time,types
from cuisine.xmlnodebase import xmlnodeclass, collectiveelement
from cuisine.desctypes import *
import logging
log=logging.getLogger("desctypes")
tlog=logging.getLogger("timing")
import xml.dom.minidom as minidom

class Channel(xmlnodeclass):
    """a channel block gives a function of one parameter (coordinate)
    in terms of a timing or position variable (value). in video files, the
    latter is usually frame numbers; in audio files, it might be
    samples or seconds.  the varying parameter can be specified in
    pieces in the same way SMIL stores animation curves. although you
    could provide the output parameter for each possible input frame
    (for example), it is preferred to use a smaller set of pieces to
    describe the curve.

    this channel object will try to keep a small number of linear
    segments for you.

    coords and values must be numeric-ish, such as Unum2 values.

    there are 
    """

    class Segment:
        x1=None # startCoord
        y1=None # startValue
        x2=None # endCoord
        y2=None # endValue
        attrmap=(('x1','startCoord'),('y1','startValue'),('x2','endCoord'),('y2','endValue'))
        def attributesfrom(self,node):
            # read attributes from xml to our data
            for a1,a2 in self.attrmap:
                setattr(self,a1,decodeunit(node.getAttribute(a2)))
        def attributesto(self,node):
            # write attributes from our data out to xml
            for a1,a2 in self.attrmap:
                node.setAttribute(a2,str(getattr(self,a1)))
        def __repr__(self):
            return "<segment x1=%r y1=%r x2=%r y2=%r>" % (self.x1,self.y1,self.x2,self.y2)
    
    def __init__(self,param):
        "param is the name of the parameter described in this channel"
        xmlnodeclass.__init__(self)
        self.param(param)

    def maximumerror(self,maxerror=None):
        
        """a channel can be designated to be 'approximate', such that
        all values are within plus-or-minus maxerror of the correct
        value. this means that a segment will not get split if you set
        a point that's not quite in line with the rest of the
        segment.

        maxerror will have to have the same units as the y coords of
        this channel. hopefully your value can serialize and
        unserialize okay.

        you cannot alter the maxerror after there are points in a
        channel, since that could result in inconsistencies (points
        that are 'close enough' actually get -forgotten-). perhaps
        someday you'll be able to -increase- the maxerror, since that
        should be ok."""

        #raise NotImplementedError("math doesn't work right yet - no not use")
        return None # <-- kills the feature

        if maxerror is not None and len(self.segments)>0:
            raise RuntimeError("cannot change a Channel's maxerror after it has segments defined")
        
        e=self._getorsetattr("maximumError",maxerror)

        # could be any unit
        e=decodeunit(e)
        
        return e

    def param(self,newval=None):
        return self._getorsetattr("param",newval)

    def readxml(self):
        # build our own list of simple segment structures
        self.segments=[] # our internal list of Segment structures
        for segnode in self.getChildrenElementsByTagName("segment"):
            try:
                seg=self.Segment()
                seg.attributesfrom(segnode)
            except Exception,e:
                log.error("malformed segment didn't have required attributes (%s)- skipping" % e)
                sys.excepthook(*sys.exc_info())
            else:
                self.segments.append(seg)
        log.debug("readxml found segments %r" % self.segments)
        
    def syncxml(self):
        # simply replace all the segments
        for segnode in self.getChildrenElementsByTagName("segment"):
            self.removeChild(segnode)
        for seg in self.segments:
            segnode=minidom.Element("segment")
            seg.attributesto(segnode)
            segnode.ownerDocument=self.ownerDocument # see xmlnodebase for more
            self.appendChild(segnode)

    def dumpallsegments(self):
        "prints the segment data (for debugging)"
        for seg in self.segments:
            print "x1=%r y1=%r x2=%r y2=%r" % (seg.x1,seg.y1,seg.x2,seg.y2)

    def setpoint(self,x,y):
        
        """adjust the y value of this x value. segments within x.step()
        of each other will be joined. (this currently works if x is to
        the right of a segment, only)

        this is the only (non-internal) method that sets points in
        segments. so, all maximumerror logic is in here, for example."""

        t=time.time()
        log.debug("setpoint %r %r " % (x,y))

        # is the coord within an existing segment?
        seg=self.findsegment(x)
        if seg:
            existingvalue=self.getsegvalue(seg,x)
            if existingvalue==y or self.maximumerror()>=abs(existingvalue-y):
                log.debug("setpoint - no change, point was correct (or within range)")
            else:
                # need to split the segment
                raise NotImplementedError
            return

        tlog.debug("checked existing segments %.01f" % ((time.time()-t)*1000))
        
        # is there a segment we can extend to the right?
        seg=self.findsegment(x-x.step())
        if seg:
            # yes there's a segment, can it be extended?
            x1,y1,x2,y2=seg.x1,seg.y1,seg.x2,seg.y2

            def comparefractions(num1,den1,num2,den2):
                """returns num1/den1==num2/den2, but works with units"""
                # num and den have different units
                if isinstance(num2,Unum2) and isinstance(num1,Unum2):
                    num2=num2.getas(num1._unit)
                    num1=num1._val
                if isinstance(den2,Unum2) and isinstance(den1,Unum2):
                    den2=den2.getas(den1._unit)
                    den1=den1._val
                return num1/den1==num2/den2

            # slope is kept as rise/run, to preserve the units
            rise=y2-y1
            run=x2-x1 # could be 0

            extendok=0
            if x1==x2: extendok=1 # point segment (can always be extended)
            elif comparefractions(rise,run,(y-y1),(x-x1)): extendok=1 # perfect fit to the line
            elif self.maximumerror(): # does the point fit within tolerance?

                numberofruns=(x-x1)/run # unitless- how many multiples of 'run' is it from x1..x
                projectedy=y1 + rise*numberofruns # this is what the straight-line y would be, at x
                
                if self.maximumerror()>=abs(projectedy-y):
                    extendok=1
                    log.info("in tolerance, projy=%s, y=%s, lt %s, diff=%s, test=%s" %
                             (projectedy,y,self.maximumerror(),abs(projectedy-y),
                              self.maximumerror()>=abs(projectedy-y)))
                    #y=projectedy # keep the line going the same way! don't let it drift with errors
                else:
                    log.info("beyond tolerance, projy=%s, y=%s, lt %s, diff=%s, test=%s" %
                             (projectedy,y,self.maximumerror(),abs(projectedy-y),
                              self.maximumerror()>=abs(projectedy-y)))

                
            if extendok:
                # ok to extend to the right
                seg.x2=x
                seg.y2=y
                log.debug("extended a segment %.01fms" % ((time.time()-t)*1000))
                return
            log.debug("setpoint - couldn't extend right, slope was %s/%s"%(rise,run))

            
        # also check for extending a segment to the left:
        # ...
                
        log.debug("setpoint - making new segment")
        seg=self.addsegment()   
        seg.x1=seg.x2=x
        seg.y1=seg.y2=y

        tlog.debug("made new segment %.01f" % ((time.time()-t)*1000))

    def getchannelvalue(self,x):
        "finds the segment for coord x and returns the y value. else returns None"
        seg=self.findsegment(x)
        if seg is None:
            return None
        return self.getsegvalue(seg,x)

    def getsegvalue(self,s,x):
        "returns the given segment's y value"

        # lines only, for now
        if not s.x1<=x<=s.x2:
            raise ValueError("coordinate is not within segment")
        if s.x1==s.x2:
            return s.y1
        t1=(x-s.x1)
        t2=(s.x2-s.x1)
        t3=t1/t2
        ret=s.y1+(s.y2-s.y1)*t3
        return ret

    def getsegments(self):
        "returns a list of Segment objects (which have x1,y1,x2,y2). read-only, please"
        return self.segments

    ##############
    ## internal
    def findsegment(self,coord):
        "returns the segment that contains the coord, or None"
        for seg in self.segments:
            if coord>=seg.x1 and coord<=seg.x2:
                return seg
        return None
        
    def addsegment(self,splinetype=None):
        seg=self.Segment()
        self.segments.append(seg)
        return seg
        

class Channels(collectiveelement):
    def childtype(self):
        return Channel
    def findparam(self,param):
        """returns a Channel of the specified param (returns just the first one found) or None"""
        found=[c for c in self.getall() if c.param()==param]
        if len(found)>0:
            return found[0]
        else:
            return None
    def byparamname(self,param):
        "gets, sets, or creates the channel node with the given parameter name."
        chans=self.getall()
        found=[c for c in chans if c.param()==param]
        if len(found)==0:
            # make new channel
            newchan=Channel(param)
            self.add(newchan)
            return newchan
        elif len(found)==1:
            # found it
            return found[0]
        else:
            # multiple matches
            log.warn("found %i matches for channel '%s' - returning first match" % (len(found),param))
            return found[0]
        
    def allparams(self):
        "returns a list of the parameters that are defined in this block"
        return [x.param() for x in self.getall()]
