from __future__ import division
import os, mmap, time
import logging

log=logging.getLogger("dvfile")
try:
    import avifile
except ImportError:
    log.info("avifile module not found - avi files will not be supported")
    
"""
these just retrieve frames from files. framepool does app-level caching.

the frames are (probably) python strings of data, or maybe read-only
buffer()-ish objects.

we assume that if we open a file for reading, it will not change size.

todo:

make these classes so they can use already-open files gracefully
(especially getting rid of the stdin hack)

add append-mode

arrange to give the reader method a Frame for it to read into, so it doesn't
make a new one each time

"""

if 0:
    class str_with_attrs(str):
        def __repr__(self):
            return "<str_with_attrs string of length %i>" % len(self)
else:
    def str_with_attrs(s):
        # skip this, i changed my mind
        return s


class Dvfile:
    """base class for disk files that hold dv frames"""
    def __init__(self,filename,dir="input",append=0):
        if dir=="input":
            self._openread(filename)
            self.__dict__['__del__']=self.closeread
        else:
            self._openwrite(filename,append)
            self.__dict__['__del__']=self.closewrite
        # perhaps we'll add 'append' someday
    def _openread(self,filename):
        raise NotImplementedError
    def _openwrite(self,filename,append):
        raise NotImplementedError
    def frames(self):
        "return number of frames in the file (..so far, if this it's being written)"
        raise NotImplementedError
    def seek(self,framenum=None):
        """if framenum isn't None, seek to that frame so it's the next
        one read.  returns current frame number, which may not be the
        input if the seek failed."""
        raise NotImplementedError
    def readframe(self):
        "get next frame and advance. returns dvrame object. throws EOFError if EOF"
        raise NotImplementedError
    def writeframe(self,frame):
        "write frame (string of data) to the end of the file"
        raise NotImplementedError
    def closewrite(self):
        raise NotImplementedError
    def closeread(self):
        raise NotImplementedError

class Dv(Dvfile):
    def _openread(self,filename):
        if filename is not None:
            self.f=open(filename,"rb")
            #            filelen=os.fstat(self.f.fileno())[6]
            #            self.mm=mmap.mmap(self.f.fileno(),filelen,prot=mmap.PROT_READ)
        else:
            self.f=sys.stdin
        self.numframes=None
        self.pos=0 # file position, to avoid tell()
        return self
    def _openwrite(self,filename,append):
        mode="wb"
        if append:
            mode="ab"
        self.f=open(filename,mode)
        # the following would be wrong in append mode
        self.numframes=0
        self.pos=0
        return self
    def readframe(self):
        #        data=self.mm[self.pos:self.pos+120000]
        data=self.f.read(120000)
        # this turns a short frame at the end (dvgrab :() into a
        # full-length-but-corrupt frame, which makes libdv not segv
        if 0<len(data)<120000: 
            data=data+"\x00"*(120000-len(data))
        frame=str_with_attrs(data)

        self.pos+=120000
        if frame=="":
            raise EOFError
        return frame
    def writeframe(self,frame):
        if len(frame)!=120000: # how rude to PAL folks!
            log.error("Frame object had %i bytes of data - not saving" % len(frame))
            return None
        ret=self.f.write(str(frame))
        self.pos+=120000
        self.numframes+=1
        return ret
    def seek(self,framenum=None):
        if framenum is None:
            # "tell"
            x=self.pos/120000
            if x!=int(x):
                raise ValueError("file position %d is not on a frame boundary" % self.pos)
            return x
        framenum=int(framenum)
        
        self.f.seek(120000*framenum)

        self.pos=120000*framenum
        return framenum
    def frames(self):
        if self.numframes is None: # (this means we're open for reading)
            self.numframes=int(os.fstat(self.f.fileno())[6]/120000)
        return self.numframes
    def closewrite(self):
        self.f.close()
    def closeread(self):
        self.f.close()

class Avi1(Dvfile):
    def _openread(self,filename):
        if filename is None:
            raise ValueError("can't read AVI from stdin")
        self.f=avifile.AVIFile() # this will open avi1 or avi2
        if not self.f.Open(filename):
            raise RuntimeError("avi file open failed")
        log.info("parsing RIFF on %s..." % filename)
        t=time.time()
        self.f.ParseRIFF()
        log.info("reading avi index...")
        self.f.ReadIndex()
        self.totalframes=self.f.GetTotalFrames()
        t=time.time()-t
        log.info("...done (%s frames %.1fs %ifps)" % (self.totalframes,t,int(self.totalframes/t)))
        self.nextframe=0
        return self

    def _openwrite(self,filename,append):
        if append:
            raise RuntimeError("append not supported for avi")
        self.f=avifile.AVI1File()
        self.f.Create(filename)
        self.f.Init(avifile.AVI_NTSC,48000,avifile.AVI_LARGE_INDEX)# fixme
        return self
    def readframe(self):
        # this does not yet handle EOF (except maybe by segfaulting)
        frame=avifile.Frame()
        if self.nextframe>=self.totalframes:
            raise EOFError
        self.f.GetDVFrame(frame,self.nextframe)
        self.nextframe+=1
        return str_with_attrs(avifile.getframedata(frame))
    def seek(self,framenum):
        self.nextframe=framenum
    def frames(self):
        return self.totalframes
    def writeframe(self,frame):
        aviframe=avifile.Frame()
        avifile.copydatafromstring(aviframe,frame)
        self.f.WriteFrame(aviframe)
    def closewrite(self):
        log.info("writing RIFF")
        self.f.WriteRIFF()
        self.f.Close()
        log.info("file closed")
    def closeread(self):
        self.f.Close()
        

class Avi2(Avi1):
    def _openwrite(self,filename):
        self.f=avifile.AVI2File()
        self.f.Create(filename)
        self.f.Init(avifile.AVI_NTSC,48000,avifile.AVI_LARGE_INDEX)
        return self


def lookupextension(format):
    return {#format:extension
              'dv1':'avi',
              'dv2':'avi',
              'raw':'dv'}[format]
def lookupformat(extensionOrFilename):
    if extensionOrFilename.endswith("avi"):
        return "dv1"
    else:
        return "raw"
def lookupformattype(format):
    return {'dv1':Avi1,'dv2':Avi2,'raw':Dv}[format]

#
# here are the factory functions that you should use to get Dvfile
# objects
#
def openForRead(filename):
    """open filename, detect its format, and return the right Dvfile
    subclass for the file."""
    f=open(filename,"rb")
    head=f.read(4)
    f.close()
    if head=="RIFF":
        return Avi1(filename,"input")
    else:
        return Dv(filename,"input")

def openForAppend(filename):
    """open filename, detect its format, and return the right
    Dvfile subclass with the file open for appending"""
    raise NotImplementedError

def openForWrite(filename,format="detect from filename",append=0):
    """open filename for writing using the given format in
    ('dv1','dv2','raw').  if format is omitted, the format is guessed
    from the file's extension. 'avi' becomes 'dv1', anything else
    becomes 'raw'"""
    if format=="detect from filename":
        format=lookupformat(filename)

    log.debug("dvfile chooses a '%s' object" % lookupformattype(format).__name__)
    return (lookupformattype(format))(filename,dir="output",append=append)
