"""for use by the desc nodes only, this is a app-wide cache of
frames/blocks read from files. when you request frames (through the
desc files), they'll come from here. this cache will keep frames
around and might even read ahead

in 2003/8, we found that knead outputted wrong movies (repeats and
skips) unless readahead was turned off. that means framepool has bugs.
"""

from __future__ import division
import sys
from cuisine import dvfile
from time import time,sleep
from sets import Set
import thread,threading
import logging
log=logging.getLogger("framepool")

cachemegsperfile=50 # no file will keep more than this much in ram 
cachemegstotal=300 # we dump entire files to attempt to maintain this amount

framelimitperfile=int(cachemegsperfile/.120000)
framelimittotal=int(cachemegstotal/.120000)

readaheadframes=[60] # a mutable integer, number of frames to read ahead on each read request

global_diskactive=0 # tells if a bkg thread is reading ahead on the disk. doesnt have to be perfect.

class Openedfile:
    def __init__(self,filename):
        self.f=dvfile.openForRead(filename)
        self.filename=filename
        self.frames={} # frame number : [frame object, load time, last accessed time]
        
        self.stopfetch=threading.Event()
        self.fetchready=threading.Event()
        self.fetchready.set()

        self.requestedframeready=threading.Condition()
        self.recentrequests=[] # list of requested framenumbers
    def __str__(self):
        return "<opened file %s, %i frames in cache>" % (self.filename,len(self.frames))
    __repr__=__str__

    def timereport(self,msg,starttime):
        log.debug(msg,(time()-starttime)*1000)

    def recentuse(self):
        """returns the access time of the most recently
        accessed open frame from this file or 0 for no frames"""
        allaccs=[lastacc for fo,lt,lastacc in self.frames.values()]
        if allaccs:
            return max(allaccs)
        else:
            return 0
    def framesincache(self):
        return len(self.frames)

    def purgeall(self):
        """dumps all the frames we're holding for this file"""
        log.info("purging all %i cached frames for %s" % (self.framesincache(),self.filename))
        self.frames.clear()

    def purgeframes(self):
        """deletes the oldest frames in our cache to leave no more
        than cachemegsperfile of frame data. the current
        implementation is ridiculous (but at least it's not called
        very often, and only in the background)."""
        t=time()
        oldest={} # last acc : framenum
        for num,row in self.frames.items():
            oldest[row[2]]=num
        oldesttimes=oldest.keys()
        if len(oldesttimes)>framelimitperfile:
            oldesttimes.sort()
            tooold=oldesttimes[:-framelimitperfile]
            log.debug("purge %i cached frames, %i are too old" % (len(self.frames.keys()),
                                                                  len(tooold)))
            for n in tooold:
                del self.frames[oldest[n]]
        self.timereport("purge took %.1fms",t)


    def purgeframes_experimental(self):
        """an alternate purge algorithm, not working. change the method names and
        enable the recentrequests lines to turn it on"""
        t=time()
        recentlyused={}  # framenum : None
        nrecentlyused=0
        recreq=self.recentrequests
        recreq.reverse()
        # count backwards all the unique recent frames
        for f,i in zip(recreq,range(len(recreq)-1,-1,-1)):
            if f not in recentlyused:
                recentlyused[f]=None
                nrecentlyused+=1
            if nrecentlyused>framelimitperfile:
                break
        # if there are still more frames in the request list, dump them
        # (this fails to dump frames that were read ahead)
        if nrecentlyused>framelimitperfile:
            dump=self.recentrequests[:i]
            for f in dump:
                del self.frames[f]
            self.recentrequests[0:i]=[]
            log.debug("recreq now %i, dumped %i" % (len(self.recentrequests),len(dump)))
        self.timereport("purge2 took %.1fms",t)

    def getoneframe(self,framenum):
        """reads the next frame from the file, puts it in cache"""
        fr=self.f.readframe()

        # store the frame in our Openedfile cache
        t=time()
        self.frames[framenum]=[fr,t,t]

    def _readframes(self,framenum,requestedframeready,stopfetch):
        """
        requestedframeready is a threading.Condition that we call
        notify on to alert getframes() that the frame it wanted is
        ready (even though we may continue reading other frames)

        """
        global global_diskactive
        log.debug("  /-- background frame fetch")

        framenum = int(framenum)

        global_diskactive=1
        requestedframeready.acquire()

        t=time()
        self.f.seek(framenum)
        self.timereport("  |   seek took %.1fms",t)
        
        self.intendedreads=(framenum,framenum+readaheadframes[0])
        for curnum in range(*self.intendedreads):

            isrequestedframe=curnum==framenum
            
            if curnum>framenum+5 and stopfetch.isSet():
                log.debug("  |   background aborted early after %i reads" % (curnum-framenum))
                stopfetch.clear()
                break
            if curnum in self.frames:
                # we've read ahead into a frame that's already in the cache. stop.
                log.debug("  |   got to a frame that was already in cache after %i reads" % (curnum-framenum))
                break

            try:
                self.getoneframe(curnum)
            except EOFError:
                if isrequestedframe:
                    log.error("the requested frame %s was past the EOF!" % curnum)
                    # tell getframe() that the requested frame is "ready"
                    requestedframeready.notify()
                    requestedframeready.release()
                    requestedframeready=None
                break
    
            if requestedframeready:
                self.timereport("  |   requested frame found in %.1fms- getframe can return it now",t)
                requestedframeready.notify()
                requestedframeready.release()
                requestedframeready=None
                
        if requestedframeready:
            log.debug("  |   left readframes before the first frame was read")
            requestedframeready.release()

        if curnum==self.intendedreads[1]-1: # only purge if we got to the end
            self.purgeframes()
        self.fetchready.set()
        log.debug("  \\-- background fetch complete")
        global_diskactive=0


    def returncachedframe(self,framenum):
        match=self.frames[framenum]
        match[2]=time()
        return match[0]
        

    def getframe(self,framenum):
        """
        this is the main function for openedfile. it takes a frame
        request and does one of these:
        
          - finds the frame in the cache and returns it

          - notices that a background process is about to read the
             frame into the cache, waits for that to happen, and
             returns the frane from the cache

          - notices that there's a background process that is *not*
             about to read the frame, tells that process to abort, and
             starts a new background process that will read the
             requested frame first, lets the process get past the
             first frame, and returns the frame

          - finds no background process, starts one which will get the
             requested frame first, lets the process get past the
             first frame, and returns the frame

        If fetching the frame caused an EOFError or similar, None will
        be returned.

        This is flawed because each file should not have its own
        background thread. there should be one smart file accessor
        that handles every file's priorities (and knows about the
        disk(s) limitations)

        """
        #self.recentrequests.append(framenum) # for one of the purge algorithms

        # start/end frames are held
        framenum = max(0,min(self.f.frames()-1,framenum))
        
        # check for cache hit
        if framenum in self.frames:
            return self.returncachedframe(framenum)


        # we have a cache miss which means we need to get
        # frames. we're going to spawn another thread to do so, and
        # that thread will readahead some number of frames. but if
        # we're seeking around a lot, the readahead might be
        # detrimental. so there's a mechanism to abort a reader thread
        # that's just doing readahead.

        log.debug("getframe cache miss on frame %i" % framenum)

        # special case - no background thread at all
       
        if readaheadframes[0]<2:
            self.f.seek(framenum) # the background thread would have done this before starting its loop
            try:
                self.getoneframe(framenum)
            except EOFError:
                return None
            return self.returncachedframe(framenum)

        # see if there's a running thread:
        if not self.fetchready.isSet():
            # maybe the requested frame is about to be read by the running thread
            if self.intendedreads[0]<=framenum<self.intendedreads[1]:
                # the thread is going to read our frame soon; better just to wait for it
                t=time()
                while framenum not in self.frames and time()<t+.2:
                    sleep(.001)
                if framenum in self.frames:
                    log.debug("frame found by running thread")
                    return self.returncachedframe(framenum)
            
            # there's a fetch in progress
            log.debug("stopping a fetch in progress")
            self.stopfetch.set() # tell it to abort early
            self.fetchready.wait(.5) # wait for it to be done
            if not self.fetchready.isSet():
                # timed out
                log.warn("timed out trying to stop a background thread - I guess there wasn't one")
                self.stopfetch.clear()
        else:
            # we dont have a running thread, but someone else might:
            #
            # that  condition is a cheap attempt to reduce disk
            # contention.  if we were unlucky enough that two preloads
            # would have happened at the same time, this may (not
            # guaranteed) make the second one just get its one frame and
            # abort, letting the first preload finish in (more) peace.
            if global_diskactive:
                log.debug("disk was busy - not preloading")
                self.f.seek(framenum)
                try:
                    self.getoneframe(framenum)
                except EOFError:
                    return None
                return self.returncachedframe(framenum)

        # there's a lock that we use to wait for the frame we actually
        # wanted (even though the thread will keep reading more frames
        # afterwards)
        self.requestedframeready.acquire()

        # we now set a flag that says a thread is running, so we can
        # detect it in later calls
        self.fetchready.clear()
        # the thread is spawned:
        thread.start_new_thread(self._readframes,(framenum,self.requestedframeready,self.stopfetch))
        # we wait for the frame that we needed to be ready
        t=time()
        self.requestedframeready.wait(1)
        self.timereport("waited %.1fms for the requested frame",t)
        self.requestedframeready.release()
        # ..and return that frame. the thread will keep going and retrieve other frames
        if framenum in self.frames:
            return self.frames[framenum][0]
        else:
            # we weren't able to come up with the requested frame (probably becuase of EOF)
            return None 


        
openfiles={} # filename : object
failedfiles = Set() # filename : exc_text

def setframereadahead(frames):
    """adjust the frame readahead - see whip for an example"""
    log.info("setting framepool readahead to %i" % frames)
    readaheadframes[0]=frames

_framesfetched=0
def masterpurge():
    """this asks each open file when it was last accessed, and purges
    the oldest files to satisfy framelimittotal (which is a function
    of cachemegstotal). we could look at each open frame and toss the
    LRU frame-by-frame, but we currently don't."""
    global openfiles
    log.debug("purging opened files")
    uses=[(v,v.recentuse()) for k,v in openfiles.items()]
    uses.sort(lambda x,y: cmp(x[1],y[1]))
    # uses is now sorted from oldest use to newest use
    acc=0
    for i in range(len(uses)-1,-1,-1):
        fic=uses[i][0].framesincache()
        acc+=fic
        if acc>framelimittotal:
            log.info("flushing file %s",uses[i][0])
            acc-=fic
            uses[i][0].purgeall()
    

def fetchframe(filename,framenum):
    """this returns a read Frame or None if the frame can't be loaded (out of range or something)"""
    global _framesfetched,openfiles

    _framesfetched+=1
    if _framesfetched%500==0:
        masterpurge()
        
    if filename in openfiles:
        # at least the file was open
        openedfile=openfiles[filename]
    else:
        # new file
        
        try:
            openedfile=Openedfile(filename)
        except IOError, e:
            if (filename,str(e)) not in failedfiles:
                failedfiles.add((filename,str(e)))
                print e
            return None
            
        openfiles[filename]=openedfile


    return openedfile.getframe(framenum)


if __name__=='__main__':
    print getframe("baseball.dv",0)
    print getframe("baseball.dv",1)
    print getframe("baseball.dv",0)
    # testing complete :)
    
    
