#include "Python.h"

#include <libdv/dv.h>
#define MAX_CHANNELS 4
#define BYTESPERSAMPLE 2

  PyObject *makeaudiobuffers( int channels, int maxsamplesperchannel ) {
    /*
      returns a python list of python buffer objects, 1 per channel.
      each buffer has enough bytes to hold maxsamplesperchannel
      samples. these are read-write buffer objects, which you can't
      easily create directly in python
    */
    PyObject *list,*buf;
    int i, bufferlength = maxsamplesperchannel * BYTESPERSAMPLE;

    if( !( list=PyList_New(channels) ) ) {
      PyErr_Format(PyExc_RuntimeError, 
		   "couldn't create list of length %d", channels);
      return NULL;
    }
    
    for( i=0; i<channels; i++ ) {
      if( !(buf = PyBuffer_New(bufferlength)) ) {
	PyErr_Format(PyExc_RuntimeError, 
		     "couldn't create new buffer object with length %d", 
		     bufferlength);
	// need to free the list somehow?
	return NULL;
      }
      
      PyList_SET_ITEM(list, i, buf);
    }
    return list;
  }



  PyObject *decodeaudio( dv_decoder_t *dv, uint8_t *buffer, 
			 PyObject *audiobufs ) {
    /*
      calls dv_decode_full_audio on the dv object and directs the
      output into the given buffers (which you made with
      makeaudiobuffers, above). 

      returns the number of samples in the frame (although that wasn't
      calculated here)

      32k audio untested
    */
    int16_t *outbufs[MAX_CHANNELS];
    void *vptr;
    int i,len,chans;
    PyObject *buf;

    chans = dv_get_num_channels(dv);
    
    //
    // set the elements of outbufs from the objs in the audiobufs list
    //
    for(i=0; i<chans; i++ ) {

      if( !(buf = PyList_GetItem(audiobufs, i)) ) { 
	return PyErr_Format(PyExc_ValueError, 
			    "buffer list element %d was null", i);
      }
      
      if( PyObject_AsWriteBuffer(buf, &vptr, &len) ) {
	return PyErr_Format(PyExc_TypeError, "couldn't get writable buffer "
			    "space from list element %d", i);
      }

      outbufs[i] = (int16_t *)vptr;

      if( len<dv_get_num_samples(dv)*BYTESPERSAMPLE ) {
	return PyErr_Format(PyExc_RuntimeError, "buffer length %d is not "
			    "long enough for %d samples", len, 
			    dv->audio->samples_this_frame);
      }
    }

    if( !dv_decode_full_audio(dv, buffer, outbufs)) {
      return PyErr_Format(PyExc_RuntimeError, "dv_decode_full_audio failed "
			  "(and probably said 'no audio'");
    }

    // at this point we could interleave the stereo making it easier
    // to handle for player.py and the mixer code (since most mixer
    // code will work fine with interleaved samples)

    return PyInt_FromLong(dv_get_num_samples(dv));
  }










#if 0


%module libdv

%include typemaps.i
%include carrays.i

/*
a good example of a typemap, but i'm not going to use it in the python api, i think

%typemap(python,in) int *pitches(int temp[3]) {
  if( PyTuple_Check($input) ) {
    if( !PyArg_ParseTuple($input,"iii",temp,temp+1,temp+2)) {
      PyErr_SetString(PyExc_TypeError, "pitch tuple must have 3 elements");
      return NULL;
    }
    $1 = (int *)temp;
    printf("input pitches are %i,%i,%i\n",temp[0],temp[1],temp[2]);
  } else {
    PyErr_SetString(PyExc_TypeError, "must pass a 3-element tuple of pitches");
    return NULL;
  }
}
*/

// where libdv takes the compressed buffer (as uint8_t), we accept a python string
%typemap(python,in) uint8_t *buffer {
  char *ptr;

  // receive python string, and get its buffer
  if (!PyString_Check($input)) {
    PyErr_SetString(PyExc_TypeError, "buffer input must be a string object");
    return NULL;
  }
#ifdef DEBUG
  fprintf(stderr,"input string to create from py object at %p\n",$input);
  fprintf(stderr,"the py str length is %i\n",PyString_Size($input));
#endif
  $1 = (unsigned char *)((void *)PyString_AS_STRING((PyStringObject *)$input));
#ifdef DEBUG
  fprintf(stderr,"python buf was at %p\n",$1);
#endif
}

/* dv_set_error_log wants a FILE* */
%typemap(in) FILE * {
  $1 = PyFile_AsFile($input);
}

/*
this arranges for all dv_color_space_t arguments in C to accept simple
strings in python. the accepted strings will be the unique parts of
the unum names.
*/
%typemap(python,in) dv_color_space_t {
  char *s;

  if( !PyString_Check($input)) {
      PyErr_SetString(PyExc_TypeError, "must pass in a color_space enum name in 'yuv','rgb','bgr0'");
      return NULL;
  }

  s=PyString_AsString($input);
  if( !strcmp(s,"yuv") ) {
    $1 = (dv_color_space_t)e_dv_color_yuv;
  } else if( !strcmp(s,"rgb")) {
    $1 = (dv_color_space_t)e_dv_color_rgb;
  } else if( !strcmp(s,"bgr0")) {
    $1 = (dv_color_space_t)e_dv_color_bgr0;
  } else {
      PyErr_SetString(PyExc_TypeError, "must pass in a color_space enum name in 'yuv','rgb','bgr0'");
      return NULL;
  }
}

/*
 dv_get_timestamp_int takes an array of 4 ints which it writes
 to. this makes a new SWIG entity that we can use in python to pass a
 valid int array to dv_get_timestamp_int.
*/
%array_class(int, intArray);


%{
  //  #define DEBUG 1
%}

/*
this gets us more complete access to dv_audio_t, dv_video_t,
dv_decoder_t, dv_encoder_t, and maybe some others.

i think these structs are on their way to being replaced by accessor
functions (which may be available already). so someday, this types.h
will probably go away
*/
%include "libdv_selected_types.h"

/*
this include brings in all the "main" dv_* functions (encoder and
decoder).
*/
%include "libdv/dv.h"

%inline %{

  /*
    use this instead of dv_get_recording_datetime_tm since it handles
    the return value correctly. i coudlnt figure out how to do this
    with a typemap.
  */
  PyObject *getrecorddate( dv_decoder_t *dv ) {
    char *output;
    struct tm ret;

    output=(char *)malloc(30*sizeof(char));
    if(output==NULL) {
      return NULL;
    }

    if(dv_get_recording_datetime_tm( dv, &ret)==0) {
      PyErr_Format(PyExc_RuntimeError, "dv_get_recording_datetime_tm failed");
      return NULL;
    }

    strftime(output,29,"%Y-%m-%dT%H:%M:%S%z",&ret);
    return PyString_FromString(output);
  }


   /*
   // from dvgrab's frame.cc, where the comment says:

   /** checks whether this frame is the first in a new recording

     To determine this, the function looks at the recStartPoint bit in
     AAUX pack 51.

     \return true if this frame is the start of a new recording 
   */
   int isnewrecording( dv_decoder_t *dv, uint8_t *buffer ) {
#if 1
     // copy from dvgrab's frame.cc 
      return (dv->audio->aaux_asc.pc2.rec_st == 0);
#else
      // call the more official looking libdv function, which is only in very fresh libdv
      return dv_is_new_recording(dv,buffer);
#endif
  }

  PyObject *encodeframe( dv_encoder_t *dv, 
			 dv_color_space_t colorspace /*(typemapped from a py string)*/, 
			 unsigned char *inputdata /* use swigbuf module to make this in python */
			 ) { 
    /* reads your data from a single buffer and returns a python
       string with the compressed frame */

    unsigned char *pixels[3];
    char *outbuffer;
    int ret;
    PyObject *retstring;

    outbuffer=(char *)malloc(120000); // how can i make python create the string now, to avoid this?
    if(!outbuffer) {
      PyErr_Format(PyExc_RuntimeError, "couldn't get memory for new frame");
      return NULL;
    }
    
    pixels[0]=inputdata;
    pixels[1]=pixels[2]=NULL;
   
    Py_BEGIN_ALLOW_THREADS;
#ifdef DEBUG
    printf("encode start dv=%p pixels=%p pixels[0]=%p out=%p\n",dv,pixels,pixels[0],outbuffer);
#endif
    ret=dv_encode_full_frame(dv, pixels, colorspace, outbuffer);     

#ifdef DEBUG
    printf("encode end\n");
#endif
    Py_END_ALLOW_THREADS;
    if (ret!=0) {
        PyErr_Format(PyExc_RuntimeError, "dv_encode_full_frame failed");
      return NULL;
    }
    retstring=PyString_FromStringAndSize(outbuffer,120000);
    free(outbuffer);
    return retstring;

  }



 

  // because i use an internal function, we need these decls which
  // aren't in the standard includes:

  // from libdv/enc_audio_input.h
  typedef struct dv_enc_audio_info_s {
    /* stored by init (but could be used for on the fly sample
       rate changes) */
    int channels;
    int frequency;
    int bitspersample;
    int bytespersecond;
    int bytealignment;
    /* stored by load */
    int bytesperframe;
    /* big endian 12/16 bit is assumed */
    unsigned char data[DV_AUDIO_MAX_SAMPLES * 2 * 4]; /* max 48000.0 Hz PAL */
  } dv_enc_audio_info_t;
  

  // from libdv/enc_output.h
  extern int _dv_raw_insert_audio(unsigned char * frame_buf, 
				  dv_enc_audio_info_t * audio, int isPAL);

  int encodeaudio( dv_encoder_t *encoder, uint8_t *buffer, 
		   PyObject *interleavedstereosamples, int frequency )
    {
      /*
	body is copied from encode.c:dv_encode_full_audio(), which does its
	own interleaving (which i didnt want)
	
	now, shouldnt the bytes per frame VARY? my input audio at 48k has
	a samples per frame pattern of 1601,1601,1602,1602,1602. this
	function seems to put the max samples onto every frame.
	there seems to be a calculate function in libdv now that will tell us
	how many samples to put on frame n.
	
	returns 0 for success
      */
      
      /* this assumes signed 16 bit pcm audio */
	int i,j;
	dv_enc_audio_info_t audio; // see enc_audio_input.h for the declaration
	audio.channels = 2;
	audio.frequency = frequency;
	audio.bitspersample = 16;
	audio.bytealignment = 4;
	audio.bytespersecond = frequency * audio.bytealignment;
	audio.bytesperframe = audio.bytespersecond/(encoder->isPAL ? 25 : 30); /* not used */

	//should assert string is shorter than the audio.data space

	swab(PyString_AsString(interleavedstereosamples), /*from*/
	     audio.data,                                  /*to*/
	     PyString_Size(interleavedstereosamples));    /*bytes*/

	return _dv_raw_insert_audio(buffer, &audio, encoder->isPAL);
    }

%}

#endif
