import gtk
try:
    import logging
    log=logging.getLogger("varvallist")
except:
    # hopefully you're running standalone and this problem
    # will be properly handled in the test script at the end of the file
    print "couldn't import logger"

"""please see docstring for DictEditor class, which is the export for
this module"""

class DictEditorRow:
    """a row in the varval table"""
    def __init__(self,parenttable,row):
        """puts in new widgets on a new row in the table (you have to
        tell which row number)."""
        self.parent=parenttable
        self.key=gtk.Entry()
        self.oldkey=None
        self.val=gtk.Entry()
        self.allwidgets=[gtk.Button("X"),
                         gtk.Label("Variable:"),
                         self.key,
                         gtk.Label("Value:"),
                         self.val]
        
        for col,wid in zip(range(0,len(self.allwidgets)),self.allwidgets):
            wid.show()
            parenttable.attachpolicy(wid,col,row)

        #self.allwidgets[0].render_icon("gtk-close",4,"GtkButton")
        self.allwidgets[0].connect("clicked",lambda *args: self.on_closebox())
        self.key.connect("changed",lambda *args: self.on_key_changed())
        self.val.connect("changed",lambda *args: self.on_val_changed())

        self._ignorechanged=0 # disables the 'changed' signal

    def on_closebox(self,*args):
        self.parent.deleterow(self.getkey(useoldkey=1))
    def on_key_changed(self,*args):
        if self._ignorechanged:
            return
        self.parent._ignoredict=1
        if self.parent.keychanged(self.oldkey,self.getkey(),self.getval()):
            # key clash - didn't adjust dict, so leave oldkey the same

            if not hasattr(self,'origstyle'):
                log.debug("entered a clash - recolor")
                self.origstyle=self.key.get_style()
                style=self.origstyle.copy()
                style.text[gtk.STATE_NORMAL]=self.key.get_colormap().alloc_color("red")
                self.key.set_style(style)
        else:
            # ok - saved to dict
            if hasattr(self,'origstyle'):
                log.debug("revert color")
                self.key.set_style(self.origstyle)
                del self.origstyle
            self.oldkey=self.getkey()
        self.parent._ignoredict=0

    def on_val_changed(self,*args):
        if self._ignorechanged:
            return
        self.parent.valchanged(self.getkey(),self.getval())
        
    def kill(self):
        """kills this row and its widgets"""
        for w in self.allwidgets:
            w.destroy()

    def getkey(self,useoldkey=0):
        """gets the current text in the 'variable' entry box. or, if
        you ask for the old key, you'll get the last key that was
        accepted as a dictionary key (as opposed to a clashing key,
        which did not get put in the dictionary). see
        DictEditor.findkey()"""
        widgettext=self.key.get_text()
        if useoldkey:
            return self.oldkey
        else:
            return widgettext
    def getval(self):
        return self.val.get_text()
    def setkey(self,k):
        self._ignorechanged=1
        self.key.set_text(str(k))
        self.oldkey=k
        self._ignorechanged=0
    def setval(self,v):
        self._ignorechanged=1
        self.val.set_text(str(v))
        self._ignorechanged=0
    
    def isblank(self):
        "are both entry boxes empty"
        return self.getval()=="" and self.getkey()==""

class DictEditor(gtk.Table):
    """growing list of variable/value entry boxes. always has exactly
    one empty pair, which is at the bottom.

    as a fancy special trick, keys that are prefixes of each other can
    still be typed in. as you're typing the new name and you type the
    character that would make the names equal, we quietly avoid
    updating the source dictionary until your key would no longer
    clash. the text becomes gray during a clash, which is intended to
    signal the user that something's wrong with his/her key.

    it's still currently possible to corrupt the widget with a few
    simultaneous prefixes.

    values will get str()'d on the way to the entry boxes, so they'll
    get rewritten as strings in your dictionary (at an undefined
    time).
    """
    def __init__(self):
        gtk.Table.__init__(self,1,4)
        self.dict=None
        self.rows=[]
        self._ignoredict=0 # supress readdict calls

    def readdict(self,newdict=None):
        """read items from the dict  and fill in our widgets. try to
        preserve the order, put new keys at the bottom. if you pass
        newdict, that becomes the dictionary that we'll read from now
        on and write changes to."""

        if self._ignoredict:
            return
        
        log.debug("updating widgets from dict")
        if newdict is not None:
            self.dict=newdict
            
        for k,v in self.dict.items():
            r=self.findkey(k)
            if r is None:
                r=self.addrow()
                r.setkey(k)
            r.setval(v)
        self.maintainblankrow()
    ##################################################################
    def islastrow(self,row):
        return row==self.rows[-1]
            
    def addrow(self):
        """creates and returns a new row at end of table"""
        r=DictEditorRow(parenttable=self,
                        row=self.get_property("n-rows"))
        
        self.rows.append(r)
        return r
    def attachpolicy(self,wid,col,row):
        "attach with our default attach settings. no col/row spanning"
        self.attach(wid,col,col+1,row,row+1,
                    (gtk.EXPAND)|(0*gtk.SHRINK)|gtk.FILL, # xoptions
                    (gtk.EXPAND)|(0*gtk.SHRINK)|gtk.FILL) # yoptions
    def findkey(self,k):
        """returns the (first) row with this key or else None. in this
        case, we want to know the key that the row maps to in the real
        dictionary, not what the entry widget happens to say. there's
        a difference when the user is typing a key name that has
        another key as a prefix. when they keys are momentarily equal,
        we lag on updating the real dict so as not to clobber the
        shorter key."""
        for r in self.rows:
            if r.getkey(useoldkey=1)==k:
                return r
        return None
    def valchanged(self,k,v):
        "callback from a DictEditorRow"
        self.dict[k]=v
        self.maintainblankrow()
    def keychanged(self,oldkey,k,v):
        """callback from a DictEditorRow.
        changes the key only if that wouldn't replace another key.
        if it would, we return 1 and don't change the dict at all"""
        log.debug("key changed from %s to %s" % (oldkey,k))
        if k in self.dict:
            # there's already a key with the new name in the dict--
            # let's hold off on the updating
            log.debug("key clash")
            self.maintainblankrow()
            return 1
        else:
            # ok to change the key, which means
            # del'ing the old one and making a new one
            if oldkey is not None:
                del self.dict[oldkey]
            self.dict[k]=v
        self.maintainblankrow()
        return 0
    def deleterow(self,key):
        "remove a dict key and the corresponding row"
        r=self.findkey(key)
        if not self.islastrow(r):
            r.kill()
            self.rows.remove(r)
            del self.dict[key]
        self.maintainblankrow()
    def maintainblankrow(self):
        blankrownums=[]
        for r,rnum in zip(self.rows,range(0,len(self.rows))):
            if r.isblank():
                blankrownums.append(rnum)
        lastrow=len(self.rows)-1
        if lastrow not in blankrownums:
            # need a new blank at the end
            self.addrow()
        else:
            blankrownums.remove(lastrow)
        for rnum in blankrownums:
            self.rows[rnum].kill()
            self.rows.pop(rnum)
            if '' in self.dict:
                del self.dict['']

if __name__=='__main__':
    try:
        import logging
        log=logging.getLogger()
        log.setLevel(logging.DEBUG)
        h=logging.StreamHandler()
        h.setFormatter(logging.Formatter("%(levelname)-5s %(filename)s:%(lineno)d: %(message)s"))
        log.addHandler(h)
    except:
        class Dummy:
            def __getattr__(self,attr):
                return lambda *args: None
        log=Dummy()

    win=gtk.Window()
    win.connect ('destroy', lambda win: gtk.main_quit ())        

    vvl=DictEditor()
    win.add(vvl)

    d={'food':'pizza','drink':'soda'}
    vvl.readdict(d)

    def idleprint(*args):
        print d
        d['serial']=d.get('serial',0)+1
        vvl.readdict()
        log.debug(".")
        return 1
    
    win.show_all()
    gtk.timeout_add(3000,idleprint)
    gtk.main()
