
def varname(chunk):
    """tries to returns the variable name in the given format
    specifier, returns None on failure. easily fooled by corrupt
    format strings"""

    var=""
    try:
        s=iter(chunk)
        while 1:
            char=s.next()
            if char=="%":
                # % marks the start of a format specifier

                # now we copy into the new chunk until the end of the specifier
                # (algorithm is from Objects/stringobject.c)
                char=s.next()

                # step over a balanced set of parens
                if char=="(":
                    pcount=1
                    while pcount>0:
                        char=s.next()
                        var+=char
                        if char==")":
                            pcount-=1
                        elif char=="(":
                            pcount+=1
                var=var[:-1] # strip final paren
                return var

    except StopIteration:
        pass
    return None


class numericwrapper:
    "by David McClosky, dmcc@bigasterisk.com"
    def __init__(self,val):
        self.val=val
    def __int__(self):
        return int(self.val)
    def __float__(self):
        return float(self.val)


class Interpolate(str):
    """a string that does interpolation with a chance for error
    correction.  i can't override the % operator, so you have to use
    .sub() as suggested in http://www.python.org/peps/pep-0292.html

    when you create the string, we split it up into chunks such each
    format specifier is alone in a chunk. to interpolate, we
    interpolate each chunk with the entire substitution dictionary. if
    a chunk fails, we call an appropriate rescue method. in this base
    class, the rescue methods simply throw exceptions (although
    they're a lot better than the current string interpolation
    exceptions). the good times come when you subclass the rescue
    methods so they can actually rescue the interpolation.

    bugs:
      %(foo)%(bar)s is not handled quite right
    
    """
    def __init__(self,s):
        str.__init__(self,s)

        self.chunks=[""] # this will be a list of strings that, if joined, would equal the master string

        try:
            s=iter(self)
            while 1:
                char=s.next()
                if char!="%":
                    # ordinary character
                    self.chunks[-1]+=char
                else:
                    # % marks the start of a format specifier
                    # make a new chunk (unless the last one never got any chars)
                    if len(self.chunks[-1])>0:
                        self.chunks.append("")

                    # now we copy into the new chunk until the end of the specifier
                    # (algorithm is from Objects/stringobject.c)
                    self.chunks[-1]+=char
                    char=s.next()
                    self.chunks[-1]+=char
                    # step over a balanced set of parens
                    if char=="(":
                        pcount=1
                        while pcount>0:
                            char=s.next()
                            self.chunks[-1]+=char
                            if char==")":
                                pcount-=1
                            elif char=="(":
                                pcount+=1
                                
                    # now copy until the format type character comes
                    while char not in '%sriduoxXeEfgGc':
                        char=s.next()
                        self.chunks[-1]+=char

                    self.chunks.append("") # new chunk
        except StopIteration:
            pass

        # if there was a format specifier at the end,
        # we'd get an empty (harmless but ugly) chunk
        if len(self.chunks[-1])==0:
            self.chunks.pop()
                        
    def chunked(self):
        "joins the chunks together so you can see they're the same as the original string"
        return "".join(self.chunks)
        
    def sub(self,substdict):

        result=""
        for chunk in self.chunks:
            try:
                processedchunk=chunk % substdict
            except KeyError,e: # name not in substdict
                processedchunk=self.rescue_missing_key(chunk,substdict,varname(chunk),e)
            except TypeError,e: # can't convert type
                # this is actually quite a spread of errors - we might
                # want to read the error text to make a more specific decision
                processedchunk=self.rescue_bad_type(chunk,substdict,varname(chunk),e)
            except ValueError,e: # format string is bad
                processedchunk=self.rescue_error(chunk,substdict,varname(chunk),e)
            except OverflowError,e: # int or float went on too long
                processedchunk=self.rescue_error(chunk,substdict,varname(chunk),e)
            except Exception,e:
                processedchunk=self.rescue_error(chunk,substdict,varname(chunk),e)
                
            result+=processedchunk
    
        return result

    """
    These are the methods you might override with your own
    handler. The caller was trying to do this:
    
      chunk % substdict

    but something was wrong with the string or the dict, so a rescue
    method was called to compute the resulting string. You can decide
    to raise the exception again (this object will not trap it
    further) or you can calculate the string some other way, and
    return that.

    Arguments are:
     chunk: a piece of format string, such as "%(foo)s"

     substdict: the entire dictionary being used to interpolate the string

     varname: the variable name (or None if unknown) that this chunk wants

     exception: the exception object we got when we tried to do the interpolation

    Barring any large syntax errors in the format string itself,
    'chunk' should contain exactly one format specifier and no other
    characters. It is mostly safe to look at the last character in the
    chunk to learn the format type, etc. Perhaps a future version will
    actually guarantee that chunk is a single, complete format
    specifier.

    The default versions of the methods don't do anything beyond what
    normal interpolation would do (i.e. raise exceptions in case of
    trouble) except they will try failed types with int() or float()
    conversions.
    """
    def rescue_error(self,chunk,substdict,varname,exception):
        raise exception
    def rescue_missing_key(self,chunk,substdict,varname,exception):
        raise exception
    def rescue_bad_type(self,chunk,substdict,varname,exception):
        return chunk % {varname:numericwrapper(substdict[varname])}        

class GentleInterpolate(Interpolate):
    """this subclass will never make exception on an interpolation
    problem; it would rather leave chunks out entirely. it will
    callback a logger function use a logger if one is provided."""

    def __new__(cls, arg, logfunc=None):
        return str.__new__(cls, arg)
    
    def __init__(self,s,logfunc=None):
        Interpolate.__init__(self,s)
        if logfunc:
            self.logfunc=logfunc
        else:
            self.logfunc=lambda x: None

    def rescue_error(self,chunk,substdict,varname,exception):
        self.logfunc(str(exception))
        return ""
    def rescue_missing_key(self,chunk,substdict,varname,exception):
        self.logfunc("missing key '%s'" % varname)
        return ""
    def rescue_bad_type(self,chunk,substdict,varname,exception):
        try:
            return super(GentleInterpolate,self).rescue_bad_type(chunk,substdict,varname,exception)
        except:
            self.logfunc("bad type for '%s': %s" % (varname,exception))
            return ""

if __name__=='__main__':
    istr=Interpolate("drew %(x)i %% %(likes)i%(python)i")
    print "|".join(istr.chunks)
    print istr.chunked()
    print istr.sub({'x':3,'likes':"4",'python':5})

    def log(x):
        print "--log-- "+x
    print GentleInterpolate("bad %(key)s %(value)i",log).sub({'key-missing':1,'value':'15x'})
