from __future__ import nested_scopes
import logging
log=logging.getLogger("templatedname")

import dispatcher
from cuisine.pour.interpolate import GentleInterpolate

class TemplatedName(dict):
    """TemplatedName is a dictionary of variablename:values that you might
    want to use in a filename.  simply get/set your variables with the
    usual dict interface, use choosetemplate() to pick one of the
    templates, and the str() the templatedname to get an interpolated name.

    the templates are just strings using python's own named foratter
    system.

    some of the variables have special names:

      var = the 'base' name, given on the command line, etc
      rec(year,month,day,hour,minute,second) = the components of the recording date

    """
    nametemplates=None # (to be) a dict of { templatename : template }

    def __init__(self):
        #
        # start up with some default templates and variables
        #
        dict.__init__(self)

        # this log channel will be given to the GentlyInterpolate
        # object so it has an output to complain about bad names
        nlog=logging.getLogger("nameinterpolation") 
        
        self.templates={}
        for k,v in ( # the two dvgrab name templates
            ("dvgrab standard", "%(path)s%(var)s%(serial)03i.%(extension)s"),
            ("dvgrab timestamped","%(path)s%(var)s%(recyear)i.%(recmonth)02i.%(recday)02i_%(rechour)02i-%(recminute)02i-%(recsecond)02i.%(extension)s"),
            ('tape-scene',"%(path)stape%(tape)s-%(scene)s-%(serial)03i.%(extension)s"),
            ):
            self.templates[k]=GentleInterpolate(v,nlog.info)
        self.choosetemplate("dvgrab standard")

        dict.update(self,{ # default vars to use in filename templates
            'var':"clip", # a default base name
            'serial':1, # number which increases on each new file split
            })

        if 0: # this doesn't work right for dict, see the manual decls below
            # we now override all the dict-changing methods with versions that also do a callback
            for methname in "__delitem__ __setitem__ clear popitem setdefault update".split():
                # get the real dict method
                dmeth=dict.__dict__[methname]
                # create an override that will call the good one and then our callback
                log.debug("cooking a replacement for %r" % (dmeth))
                cbmeth=lambda *args,**kwargs: self.meth_then_callback(dmeth,args,kwargs)
                # put the new method into this instance
                self.__dict__[methname]=cbmeth

    def meth_then_callback(self,meth,args,kwargs):
        log.debug("overriden method %r, args=%r, kw=%r"%(meth,args,kwargs))
        meth(self,*args,**kwargs)
        self.runcallback()

    # these are what the loop in __init__ was supposed to make, but there were issues
    # they cause any dictionary-changing calls to be followed with a callback so the
    # gui can update or whatever
    def __delitem__(self,*args,**kw): self.meth_then_callback(dict.__delitem__,args,kw)
    def __setitem__(self,*args,**kw): self.meth_then_callback(dict.__setitem__,args,kw)
    def clear(self,*args,**kw):       self.meth_then_callback(dict.clear,args,kw)
    def popitem(self,*args,**kw):     self.meth_then_callback(dict.popitem,args,kw)
    def setdefault(self,*args,**kw):  self.meth_then_callback(dict.setdefault,args,kw)
    def update(self,*args,**kw):      self.meth_then_callback(dict.update,args,kw)
    
    def gettemplatechoices(self):
        "returns all the names of templates"
        return self.templates.keys()
    def getvariables(self,templatename="<current>"):
        "returns all variable names needed to interpolate the given template"

        if templatename=="<current>":
            templatename=self.currenttemplatename

        class reportingdict:
            def __init__(self,vars): 
                self.vars=vars
            def __getitem__(self,k):
                self.vars.append(k)
                return 0
        vars=[]
        self.templates[templatename] % reportingdict(vars)
        return vars
        
    def runcallback(self):
        """a callback for any changes that might affect the
        outputted name including all dictionary changes and any
        template changes. this also fires the callback once."""

        log.debug("running callback")
        dispatcher.send("templatedname changed",sender=self)

    def createname(self):
        "returns an interpolated name with the current vars and template"
        try:
            return self.templates[self.currenttemplatename].sub(self)
        except:
            log.error("couldn't interpolate name: %s" % self.templates[self.currenttemplatename])
            return ""
    __str__=createname

    def choosetemplate(self,templatename):
        """sets the template, adds all required keys to the dictionary,
        and sends the template changed signal
        """
        if templatename not in self.templates:
            raise ValueError("template '%s' is unknown"%templatename)
        self.currenttemplatename=templatename
        for k in self.getvariables():
            if k not in self:
                self[k]=""
        self.runcallback()
    def templateneedsrecorddate(self):
        """returns if the template depends on recyear, recmonth, etc. if yes,
        you should set those variables before creating the name"""
        # fixme
        return self.templates[self.currenttemplatename].find("%(rec")!=-1


        
            
