from __future__ import division, nested_scopes, generators
from cuisine.xmlnodebase import xmlnodeclass,xmldocfile,collectiveelement
import sys,os,new,bisect
from dispatch import dispatcher
import xml.dom.minidom as minidom
from cuisine.desctypes import *
from cuisine.utils import *
from cuisine.external.dictmixin import DictMixin

from cuisine import shownodes

import logging
log=logging.getLogger("shownodes")

class EditBrickMixin:
    """not quite a mixin. it's some horrible child something or other.
    methods will replace the version in Brick (with no possiblity for
    calling the old ones, which is why some stuff gets factored out into
    separate methods that brick and editbrick can both use)
    """
    def cuts(self):
        """returns modifyable Duration:Track dict-like object.
        the cuts are stored as times offset from this brick's origin.
        all we know is which tracks to cut to; you have to go find out
        what clips are there separately. tracks are stored by name in
        the xml, but right here we convert to and from Track objects.

        in the xml, the cuts look like this:
        <cuts>
          <!--the cut with the lowest pos covers everything from before that cut, too.
              in this case, everything below 07:15 is track 3-->
          <cut pos='00:00:05:00' track='3'/> <!--a cut to track 3 at 5 seconds-->
          <cut pos='00:00:07:15' track='2'/>
        </cuts>

        there are still some undefined semantics for cuts that fall outside the
        intime/outtime of the brick. see cuts().cutafter() in particular.
        """
        class CutsDict(DictMixin):
            def __init__(self,brick):
                self.brick=brick
                
                # i'm not getting the <cuts> block in advance, since
                # that wouldn't be quite as lazy as doing it only when
                # the app touches the block. after the first
                # get/set/del though, this will contain the
                # minidom.Element of the <cuts>...</cuts> block. see
                # the _getcutsblock() method.                
                self._cutsblock=None
            def _getcutsblock(self):
                if not self._cutsblock:
                    self._cutsblock=self.brick._getsetorcreate1("cuts")
                return self._cutsblock
            def allcuts(self):
                "get the brick's cuts block as a list of Elements"
                cutsblock=self._getcutsblock()
                # remember to skip the extra text nodes
                ret=cutsblock.getChildrenElementsByTagName("*")
                return ret

                
            def getelementbypos(self,key):
                "internal func to find an Element for a given pos (as a Duration) - throws KeyError"
                for c in self.allcuts():
                    pos=Duration(c.getAttribute("pos"))
                    if pos==key:
                        return c
                raise KeyError("cut position %r not found in cuts list" % key)
            def __getitem__(self, key):
                L = self.brick._sortedcutslist()

                # common to search for increasing keys
                try:
                    lastpos,lastkey = self._getitem_last_find
                except AttributeError:
                    lopos=0
                else:
                    if lastkey<=key:
                        lopos=lastpos

                pos = None
                if lopos+1<len(L) and L[lopos+1][0]==key:
                    pos = lopos+1

                if pos is None:
                    pos = bisect.bisect(L,(key,None),lopos) # lo=
                if L[pos][0]==key:
                    self._getitem_last_find = (pos,key)
                    return L[pos][1]

                raise KeyError("position not in cuts list")
            
            def __setitem__(self, key, value):
                """changes the track for an existing element, or adds
                a new Element to the <cuts> group with the given
                show-time and Track"""
                if not isinstance(value,shownodes.Track):
                    raise ValueError("cuts values must be provided as Track objects")
                try:
                    cut=self.getelementbypos(key)
                except:
                    # make new element
                    log.info("making new element in edit at time %s" % key)
                    cutsblock=self._getcutsblock()
                    trackname=value.name()
                    elem=minidom.Element("cut")
                    elem.setAttribute("pos",str(key)) # store the Duration
                    elem.setAttribute("track",value.name())  # we store the track by name
                    elem.ownerDocument=cutsblock.ownerDocument
                    cutsblock.appendChild(elem)
                else:
                    # change existing element
                    cut.setAttribute("track",value.name())
                self.brick._flushcutslist()
            def __delitem__(self, key):
                cut=self.getelementbypos(key)
                cutsblock=self._getcutsblock()
                cutsblock.removeChild(cut)
                self.brick._flushcutslist()
            def keys(self):
                # bonus - these come out sorted
                return [x for x,v in self.brick._sortedcutslist()] # special list emits keys already
#            def __contains__(self,k):
#                this could use bisect to quickly find the key (or not) in the sorted list
            def items(self):
                # these items come out sorted by key
                itms=[]
                for k,v in self.brick._sortedcutslist():
                    itms.append((k,v))
                return itms
                

            #
            # now come some more exotic methods which pertain to the cuts list, but aren't
            # simple get/set methods on the list itself.
            #
            def cutafter(self,key):
                """returns the time coordinate (brick time) of the end
                of this cut, which might be the start of the next cut
                (in which case it's a valid key to this cuts() mapping
                object), or it might be None, meaning 'no cut points
                come after this one'. """
                
                time_track = self.brick._sortedcutslist()

                lopos=0
                pos=None
                if hasattr(self,'_cutafter_last'):
                    lastkey,lastpos = self._cutafter_last
                    if lastkey<key:
                        lopos=lastpos
                    if (lastpos+1<len(time_track)-1 and
                        time_track[lastpos+1][0]==key):
                        pos = lastpos+1

                if pos is None:
                    pos = bisect.bisect(time_track,(key,None),lopos) # lo=
                if time_track[pos][0]!=key:
                    raise KeyError("'%r' was not a valid cut" % key)
                if pos==len(time_track)-1:
                    # there wasn't anything afterwards
                    return None
                self._cutafter_last = key,pos
                return time_track[pos+1][0]
                
            def findcutstart(self,t,returntrackinstead=0):
                """given a time in the brick, we return the starting time of
                the cut that contains t (relative to this brick's
                origin).

                remember the rule is that the first cut names the track that
                is used before and after that cut. if you request a time
                before the first cut, you'll get the first cut. in any case,
                you'll get a valid key to the cuts() mapping (or a IndexError
                exception, see below)"""

                # first find what track is being input from at the given time:

                cuttimes = self.brick._sortedcutslist()

                justkeys=[x[0] for x in cuttimes]
                i=max(0,bisect.bisect_right(justkeys,t)-1)

                if returntrackinstead:
                    lasttrack=cuttimes[i][1]#.getvalue(i)
                    return lasttrack
                else:
                    lastcutpoint=cuttimes[i][0]
                    return lastcutpoint

        return CutsDict(self)
        if 0: # from dsawyer version- maybe not needed
            # always return the same cutsdict, so it can use its own cache
            if not hasattr(self,"_cutsdict"):
                self._cutsdict = CutsDict(self)
            return self._cutsdict

    def _sortedcutslist(self):
        """internal version of the cuts list as a time-sorted
        (time,track),(time,track),... list. times are measured from self.origin
        """
        if not hasattr(self,'cachedcutslist'):
            allcuts=self._getsetorcreate1("cuts").getChildrenElementsByTagName("*")
           
            self.cachedcutslist=[] # list of (pos,track),(pos,track).. sorted by pos
            show=self.parentbytype(shownodes.Show)
            for c in allcuts:
                pos=Duration(c.getAttribute("pos"))
                trackname=c.getAttribute("track")
                track=show.trackbyname(trackname)
                self.cachedcutslist.append( (pos,track) )
            
            self.cachedcutslist.sort(lambda x,y: cmp(x[0],y[0])) # sort by pos
            
        return self.cachedcutslist
    def _flushcutslist(self):
        "the other methods call this in case they do anything that might disturb the list"
        if hasattr(self,'cachedcutslist'):
            del self.cachedcutslist

    def trackfortime(self,t):
        """looks in the cuts list to find out which input track is effective
        at the given time (relative to this brick's origin). the Track is returned"""

        return self.cuts().findcutstart(t,returntrackinstead=1)

    def tbd_getframe(self,t,component):
        """fetches the frame at time t by calling tbd_getframe from the
        input brick for time t (t is relative to this edit brick's origin)

        """

        return self.getothertrackframe(self.trackfortime(t),t,component)
        

    def slice(self,t):
        """slice is a little different since the cuts shall go to one
        side or the other; they're not to be duplicated"""

        log.info("edit brick slice")

        show=self.parentbytype(shownodes.Show)
        self2=self.copy()
        show.bricks().add(self2)
        self.sliceadjusttimes(t,self2)

        self2.cuts().clear()

        self._flushcutslist()
        self2._flushcutslist()

        cuts1=self.cuts()
        cuts2=self2.cuts()
        for pos,track in cuts1.items():
            log.debug("examing cut %s %s" % (pos,track))
            if pos>=t:
                log.debug("..it belongs in the right half")
                # position is in self2, so we move the cut there
                log.debug("left half: %r" % cuts1)
                log.debug("right half: %r" % cuts2)
                cuts2[pos]=track
                del cuts1[pos]
                log.debug("post left half: %r" % cuts1)
                log.debug("post right half: %r\n" % cuts2)
                
        # also give self2 a starting cut
        if t not in cuts2:
            cuts2[t]=cuts1.findcutstart(t,returntrackinstead=1)

        return self2
    
        
