"""manages the layout of the timeline display- the coordinate systems for the
times and tracks.

the interesting coord systems are:

show time (Duration)

timeline window x/y (pixels) - the opengl surface deliberately has these coords too


the layout does not know anything about Tk or opengl. it does know
about the general arrangement of the timeline, trackview, and time
scale. 


"""
from __future__ import division
from math import exp
from cuisine.desctypes import Duration
from dispatch import dispatcher

class Layout(object):
    """this object keeps track of the transformation between
    time-in-show and position-in-window, as well as the window
    coordinates of the track displays.

    signals: if the view transform changes, the 'view transform
    changed' signal is sent.
    
    """
    #
    # the layout data
    #
    _timetransform_offset=0 
    def _g(self):
        return self._timetransform_offset
    def _s(self,o):
        if self._timetransform_offset!=o:
            self._timetransform_offset=o
            dispatcher.send("view transform changed")
    timetransform_offset=property(_g,_s,doc="x pixel coordinate of the show's time origin")

    _timetransform_scale=100 
    def _g(self):
        return self._timetransform_scale
    def _s(self,newscale):
        if newscale!=self._timetransform_scale:
            self._timetransform_scale=max(min(newscale,1000),.0001)
            dispatcher.send("view transform changed")
    timetransform_scale=property(_g,_s,doc="horizontal pixels per second of show")

    # list of Track,startrow tuples. the trackview exclusively
    # gets all y values until and not including the next one's
    # startrow. this list always sorted top-to-bottom (increasing
    # y). there's a dummy element off the end of this list of (None,y)
    # to keep track of the height of the bottommost trackview.
    # we might keep the Trackviews too someday.
    _trackviews=None    # for internal use only. other objs should use the methods.
    _trackviewobjs=None # internal. list of Trackview

    defaulttrackheight=100 # pixels

    def __init__(self):
        """there's probably only one layout per program"""
        self._trackviews=[(None,0)] # the bottom of the trackviews
        self._trackviewobjs=[]

    #
    # the coordinate conversion functions
    # 
    def windowy_for_track(self,track):
        
        """given a track object, returns both the top and bottom y
        window coords for the track. the trackview may use up to but
        not including the bottom row.  like, the trackview may use
        rows in range(top,bottom).

        a track that is scrolled out of the way will have identical
        top and bottom coords (like it's 'crushed' against the top or
        bottom of the window)."""
        
        top,bottom=None,None
        for tr,startrow in self._trackviews:
            if top is not None:
                bottom=startrow
                break
            if tr==track:
                top=startrow

        if top is None or bottom is None:
            raise ValueError("track %r is unknown to the layout" % track)
            
        return top,bottom

    def track_for_windowy(self,y):
        
        """given a y pixel coordinate in the window, returns the track
        object that's underneath (or None if there's no match)"""

        prevtrack=None
        for tr,startrow in self._trackviews:
            if y<startrow:
                return prevtrack
            prevtrack=tr
        return None

    def windowx_for_showtime(self,showtime):
        
        """given a Duration in the show, returns the window x
        coordinate (which may be outside of the visible window)"""

        return int(showtime.getas('seconds')*self.timetransform_scale+self.timetransform_offset)
    
    def showtime_for_windowx(self,x):

        """given an x pixel coordinate in the window, returns the
        Duration into the show (or None if the coord is not over the
        show)"""
        
        return Duration((x-self.timetransform_offset)/self.timetransform_scale,'seconds')

    #
    # adjusting the time (horizontal) transformation
    #

    def adjustzoom_start(self,startx):

        """lets a mouse drag adjust the zoom. pass the starting mouse
        coordinate now, and call adjustzoom_move with additional mouse
        coordinates"""
        self._zoomstartmousex=startx
        self._zoomstarttransform=self.timetransform_offset,self.timetransform_scale
        self._pivottime=self.showtime_for_windowx(startx)
        # yes, we'll be leaving these attributes lying around

    def adjustzoom_move(self,movedx):

        """lets a mouse drag adjust the zoom. after calling
        adjustzoom_start with the mousedown position, call this method
        with updated positions and the zoom will be altered."""

        delta=movedx-self._zoomstartmousex
        origoffset,origscale=self._zoomstarttransform
        
        # adjust zoom, but keep the time at _pivottime still

        # do the zoom part
        self.timetransform_scale=origscale*exp(delta/60)
        # see where the time went that was supposed to hold still
        drifted=self.windowx_for_showtime(self._pivottime)
        # fix up the offset so the pivot point holds still
        self.timetransform_offset=self.timetransform_offset-(drifted-self._zoomstartmousex)

    def adjustoffset_start(self,startx):

        """lets a mouse move adjust the offset. pass the starting
        mouse coordinate now, and call adjustoffset_move with more
        coords"""
        self._panlastmousex=startx

    def adjustoffset_move(self,movedx):

        """lets a mouse move adjust the offset. after calling
        adjustoffset_start with the mousedown position, call this
        method with updated positions and the time offset will be
        altered."""
        delta=movedx-self._panlastmousex
        self.timetransform_offset=self.timetransform_offset+delta
        self._panlastmousex=movedx

    def _(): pass
    #
    # methods for talking to trackviews
    #
    def newtrackview(self,trackview):
        
        """this is how a new trackview introduces itself. once
        registered, the layout will position the track at the bottom,
        and the track can now ask for its top/bottom y coordinates.

        the new track will be self.defaulttrackheight pixels tall.

        the layout will use trackview.track to find out what track in
        the show the trackview is for.
        """

        # the last _trackviews element is a dummy to keep track of the
        # bottom of the former last track. now we're the trackview in
        # that place, and we make a new dummy element to keep track of
        # our bottom.

        dummy,lowesty=self._trackviews[-1]
        self._trackviews[-1]=(trackview.track,lowesty)
        self._trackviews.append((None,lowesty+self.defaulttrackheight))
        self._trackviewobjs.append(trackview)
        
    def changetrackviewheight(self,trackview,newheight):
        """
        this -might- change the trackview height to the given
        height. it'll stop the trackview from getting too big or
        small.

        """
        offset=0
        for i in range(len(self._trackviews)):
            tr,y=self._trackviews[i]
            self._trackviews[i]=tr,y+offset
            if tr==trackview.track:
                oldheight=self._trackviews[i+1][1]-y
                offset=newheight-oldheight

        dispatcher.send("view transform changed")
        
    def alltrackviews(self):
        """returns list of all trackviews"""
        return self._trackviewobjs
